<?php namespace RVR\Controller;

use DateTime;
use Firebase\JWT\JWT;
use Firebase\JWT\Key;
use Firebase\JWT\SignatureInvalidException;
use Firebase\JWT\BeforeValidException;
use Firebase\JWT\ExpiredException;
use RVR\Repository\OAuthSessionRepository;
use RVR\Repository\OAuthTokenRepository;
use RVR\Repository\UserRepository;
use RVR\PersistentData\Model\User;
use RVR\PersistentData\Model\OAuthSession;
use RVR\PersistentData\Model\OAuthToken;
use RVR\Repository\OAuthClientRepository;
use SokoWeb\Interfaces\Response\IContent;
use SokoWeb\Response\JsonContent;

class OAuthController
{
    private OAuthClientRepository $oAuthClientRepository;

    private OAuthSessionRepository $oAuthSessionRepository;

    private OAuthTokenRepository $oAuthTokenRepository;

    private UserRepository $userRepository;

    public function __construct()
    {
        $this->oAuthClientRepository = new OAuthClientRepository();
        $this->oAuthSessionRepository = new OAuthSessionRepository();
        $this->oAuthTokenRepository = new OAuthTokenRepository();
        $this->userRepository = new UserRepository();
    }

    public function generateToken(): ?IContent
    {
        $credentials = $this->getClientCredentials();
        $code = \Container::$request->post('code');
        $redirectUri = \Container::$request->post('redirect_uri');

        if (!$credentials['clientId'] || !$code || !$redirectUri) {
            return new JsonContent([
                'error' => 'An invalid request was made.'
            ]);
        }

        $client = $this->oAuthClientRepository->getByClientId($credentials['clientId']);
        if ($client === null) {
            return new JsonContent([
                'error' => 'Client is not found.'
            ]);
        }

        $redirectUriBase = explode('?', $redirectUri)[0];
        if (!in_array($redirectUriBase, $client->getRedirectUrisArray())) {
            return new JsonContent([
                'error' => 'Redirect URI \'' . $redirectUriBase .'\' is not allowed for this client.'
            ]);
        }

        $session = $this->oAuthSessionRepository->getByCode($code);
        if ($session === null || $session->getTokenClaimed() || $session->getExpiresDate() < new DateTime()) {
            return new JsonContent([
                'error' => 'The provided code is invalid.'
            ]);
        }

        $codeChallenge = $session->getCodeChallenge();
        if ($codeChallenge === null && $client->getClientSecret() !== $credentials['clientSecret']) {
            return new JsonContent([
                'error' => 'Client is not authorized.'
            ]);
        }

        if ($session->getClientId() !== $credentials['clientId']) {
            return new JsonContent([
                'error' => 'This code cannot be used by this client!'
            ]);
        }

        if ($codeChallenge !== null) {
            $codeVerifier = \Container::$request->post('code_verifier') ?: '';
            if ($session->getCodeChallengeMethod() === 'S256') {
                $hash = rtrim(strtr(base64_encode(hash('sha256', $codeVerifier, true)), '+/', '-_'), '=');
            } else {
                $hash = $codeVerifier;
            }
            if ($codeChallenge !== $hash) {
                return new JsonContent([
                    'error' => 'Code challenge failed!'
                ]);
            }
        }

        $session->setTokenClaimed(true);
        \Container::$persistentDataManager->saveToDb($session);

        $token = new OAuthToken();
        $token->setSession($session);
        $token->setCreatedDate(new DateTime());
        $token->setExpiresDate(new DateTime('+1 hours'));
        \Container::$persistentDataManager->saveToDb($token);

        $commonPayload = [
            'iss' => $_ENV['APP_URL'],
            'iat' => $token->getCreatedDate()->getTimestamp(),
            'nbf' => $session->getCreatedDate()->getTimestamp(),
            'exp' => $token->getExpiresDate()->getTimestamp(),
            'aud' => $session->getClientId(),
            'nonce' => $session->getNonce()
        ];
        $idTokenPayload = array_merge($commonPayload, $this->getUserInfoInternal(
            $this->userRepository->getById($session->getUserId()),
            $session->getScopeArray())
        );
        $accessTokenPayload = array_merge($commonPayload, [
            'jti' => $token->getId(),
        ]);

        $privateKey = file_get_contents(ROOT . '/' . $_ENV['JWT_RSA_PRIVATE_KEY']);
        $idToken = JWT::encode($idTokenPayload, $privateKey, 'RS256', $_ENV['JWT_KEY_KID']);
        $accessToken = JWT::encode($accessTokenPayload, $privateKey, 'RS256', $_ENV['JWT_KEY_KID']);

        return new JsonContent([
            'access_token' => $accessToken,
            'expires_in' => $token->getExpiresDate()->getTimestamp() - (new DateTime())->getTimestamp(),
            'scope' => $session->getScope(),
            'id_token' => $idToken,
            'token_type' => 'Bearer'
        ]);
    }

    public function introspectToken(): ?IContent
    {
        $credentials = $this->getClientCredentials();
        $accessToken = \Container::$request->post('token');

        if (!$credentials['clientId'] || !$credentials['clientSecret'] || !$accessToken) {
            return new JsonContent([
                'error' => 'An invalid request was made.'
            ]);
        }

        $client = $this->oAuthClientRepository->getByClientId($credentials['clientId']);
        if ($client === null || $client->getClientSecret() !== $credentials['clientSecret']) {
            return new JsonContent([
                'error' => 'Client is not authorized.'
            ]);
        }

        $tokenValidated = $this->validateTokenAndSession($accessToken, $token, $session);
        if (!$tokenValidated) {
            return new JsonContent([
                'active' => false
            ]);
        }

        if ($session->getClientId() !== $credentials['clientId']) {
            return new JsonContent([
                'active' => false
            ]);
        }

        return new JsonContent([
            'active' => true,
            'scope' => $session->getScope(),
            'client_id' => $session->getClientId(),
            'exp' => $token->getExpiresDate()->getTimestamp(),
        ]);
    }

    public function revokeToken(): ?IContent
    {
        $credentials = $this->getClientCredentials();
        $accessToken = \Container::$request->post('token');

        if (!$credentials['clientId'] || !$credentials['clientSecret'] || !$accessToken) {
            return new JsonContent([
                'error' => 'An invalid request was made.'
            ]);
        }

        $client = $this->oAuthClientRepository->getByClientId($credentials['clientId']);
        if ($client === null || $client->getClientSecret() !== $credentials['clientSecret']) {
            return new JsonContent([
                'error' => 'Client is not authorized.'
            ]);
        }

        $tokenValidated = $this->validateTokenAndSession($accessToken, $token, $session);
        if (!$tokenValidated) {
            return new JsonContent([]);
        }

        $session = $this->oAuthSessionRepository->getById($token->getSessionId());
        if ($session->getClientId() !== $credentials['clientId']) {
            return new JsonContent([]);
        }

        \Container::$persistentDataManager->deleteFromDb($token);

        return new JsonContent([]);
    }

    public function getUserInfo() : IContent
    {
        $authorization = \Container::$request->header('Authorization');
        if ($authorization === null) {
            return new JsonContent([
                'error' => 'No Authorization header was sent.'
            ]);
        }

        $accessToken = substr($authorization, strlen('Bearer '));
        $tokenValidated = $this->validateTokenAndSession($accessToken, $token, $session);
        if (!$tokenValidated) {
            return new JsonContent([
                'error' => 'The provided access token is invalid.'
            ]);
        }

        return new JsonContent(
            $this->getUserInfoInternal(
                $this->userRepository->getById($session->getUserId()),
                $session->getScopeArray()
            )
        );
    }

    public function getConfig(): IContent
    {
        return new JsonContent([
            'issuer' => $_ENV['APP_URL'],
            'authorization_endpoint' => \Container::$request->getBase() . \Container::$routeCollection->getRoute('oauth.auth')->generateLink(),
            'token_endpoint' => \Container::$request->getBase() . \Container::$routeCollection->getRoute('oauth.token')->generateLink(),
            'introspection_endpoint' => \Container::$request->getBase() . \Container::$routeCollection->getRoute('oauth.token.introspect')->generateLink(),
            'revocation_endpoint' => \Container::$request->getBase() . \Container::$routeCollection->getRoute('oauth.token.revoke')->generateLink(),
            'userinfo_endpoint' => \Container::$request->getBase() . \Container::$routeCollection->getRoute('oauth.userinfo')->generateLink(),
            'end_session_endpoint' => \Container::$request->getBase() . \Container::$routeCollection->getRoute('logout')->generateLink(),
            'jwks_uri' => \Container::$request->getBase() . \Container::$routeCollection->getRoute('oauth.certs')->generateLink(),
            'response_types_supported' =>
            [
                'code',
            ],
            'subject_types_supported' =>
            [
                'public',
            ],
            'id_token_signing_alg_values_supported' =>
            [
                'RS256',
            ],
            'scopes_supported' =>
            [
                'openid',
                'email',
                'profile',
            ],
            'token_endpoint_auth_methods_supported' =>
            [
                'client_secret_basic',
                'client_secret_post',
            ],
            'claims_supported' =>
            [
                'aud',
                'email',
                'exp',
                'full_name',
                'iat',
                'id_number',
                'iss',
                'nickname',
                'phone',
                'picture',
                'sub',
                'username',
            ],
            'code_challenge_methods_supported' =>
            [
                'plain',
                'S256',
            ],
            'grant_types_supported' =>
            [
                'authorization_code',
            ],
        ]);
    }

    public function getCerts(): IContent
    {
        $publicKey = file_get_contents(ROOT . '/' . $_ENV['JWT_RSA_PUBLIC_KEY']);
        $keyInfo = openssl_pkey_get_details(openssl_pkey_get_public($publicKey));

        return new JsonContent(['keys' => [
            [
                'kty' => 'RSA',
                'alg' => 'RS256',
                'use' => 'sig',
                'kid' => $_ENV['JWT_KEY_KID'],
                'n' => str_replace(['+', '/'], ['-', '_'], base64_encode($keyInfo['rsa']['n'])),
                'e' => str_replace(['+', '/'], ['-', '_'], base64_encode($keyInfo['rsa']['e'])),
            ]
        ]]);
    }

    private function getClientCredentials(): array
    {
        $authorization = \Container::$request->header('Authorization');
        if ($authorization !== null) {
            $basicAuthEncoded = substr($authorization, strlen('Basic '));
            $basicAuth = explode(':', base64_decode($basicAuthEncoded));
            if (count($basicAuth) === 2) {
                $clientId = rawurldecode($basicAuth[0]);
                $clientSecret = rawurldecode($basicAuth[1]);
            } else {
                $clientId = null;
                $clientSecret = null;
            }
        } else {
            $clientId = \Container::$request->post('client_id');
            $clientSecret = \Container::$request->post('client_secret');
        }
        return ['clientId' => $clientId, 'clientSecret' => $clientSecret];
    }

    private function validateTokenAndSession(
        string $accessToken,
        ?OAuthToken &$token,
        ?OAuthSession &$session): bool
    {
        $publicKey = file_get_contents(ROOT . '/' . $_ENV['JWT_RSA_PUBLIC_KEY']);
        try {
            $payload = JWT::decode($accessToken, new Key($publicKey, 'RS256'));
            $token = $this->oAuthTokenRepository->getById($payload->jti);
        } catch (SignatureInvalidException | BeforeValidException | ExpiredException) {
            $token = null;
        } catch (\UnexpectedValueException $e) {
            error_log($e->getMessage() . ' Token was: ' . $accessToken);
            $token = null;
        }

        if ($token === null || $token->getExpiresDate() < new DateTime()) {
            return false;
        }

        $session = $this->oAuthSessionRepository->getById($token->getSessionId());
        if ($session === null) {
            return false;
        }

        return true;
    }

    /**
     * @param User $user
     * @param string[] $scope
     * @return array<string, string>
     */
    private function getUserInfoInternal(User $user, array $scope): array
    {
        $userInfo = [];
        if (in_array('openid', $scope)) {
            $userInfo['sub'] = (string)$user->getId();
        }
        if (in_array('email', $scope)) {
            $userInfo['email'] = $user->getEmail();
        }
        if (in_array('profile', $scope)) {
            if ($user->getUsername() !== null) {
                $userInfo['preferred_username'] = $user->getUsername();
            }
            $userInfo['name'] = $user->getFullName();
            $userInfo['nickname'] = $user->getNickname();
            $userInfo['phone_number'] = $user->getPhone();
            $userInfo['id_number'] = $user->getIdNumber();
        }
        return $userInfo;
    }
}
